/***************************************************************************
****************************************************************************

This is a part of the xModule Source code.

Copyright (C) 2003-2007 The LS-Universe Team

  Homepage: http://www.ls-universe.info

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

****************************************************************************
****************************************************************************/

#ifndef _ICON_EXTRACTOR_H
#define _ICON_EXTRACTOR_H

const int MAX_ICONS = 200;		  //Increase this value if required

#pragma pack( push )
#pragma pack( 2 )
typedef struct
{
	BYTE	bWidth;               // Width of the image
	BYTE	bHeight;              // Height of the image (times 2)
	BYTE	bColorCount;          // Number of colors in image (0 if >=8bpp)
	BYTE	bReserved;            // Reserved
	WORD	wPlanes;              // Color Planes
	WORD	wBitCount;            // Bits per pixel
	DWORD	dwBytesInRes;         // how many bytes in this resource?
	WORD	nID;                  // the ID
} MEMICONDIRENTRY, *LPMEMICONDIRENTRY;

typedef struct 
{
	WORD			idReserved;   // Reserved
	WORD			idType;       // resource type (1 for icons)
	WORD			idCount;      // how many images?
	MEMICONDIRENTRY	idEntries[1]; // the entries for each image
} MEMICONDIR, *LPMEMICONDIR;
#pragma pack( pop )

typedef struct
{
	UINT			Width, Height, Colors; // Width, Height and bpp
	LPBYTE			lpBits;                // ptr to DIB bits
	DWORD			dwNumBytes;            // how many bytes?
	LPBITMAPINFO	lpbi;                  // ptr to header
	LPBYTE			lpXOR;                 // ptr to XOR image bits
	LPBYTE			lpAND;                 // ptr to AND image bits
} ICONIMAGE, *LPICONIMAGE;

typedef struct
{
	BOOL		bHasChanged;                     // Has image changed?
	UINT		nNumImages;                      // How many images?
	ICONIMAGE	IconImages[1];                   // Image entries
} ICONRESOURCE, *LPICONRESOURCE;


// These next two structs represent how the icon information is stored
// in an ICO file.
typedef struct
{
	BYTE	bWidth;               // Width of the image
	BYTE	bHeight;              // Height of the image (times 2)
	BYTE	bColorCount;          // Number of colors in image (0 if >=8bpp)
	BYTE	bReserved;            // Reserved
	WORD	wPlanes;              // Color Planes
	WORD	wBitCount;            // Bits per pixel
	DWORD	dwBytesInRes;         // how many bytes in this resource?
	DWORD	dwImageOffset;        // where in the file is this image
} ICONDIRENTRY, *LPICONDIRENTRY;

typedef struct 
{
	WORD			idReserved;   // Reserved
	WORD			idType;       // resource type (1 for icons)
	WORD			idCount;      // how many images?
	ICONDIRENTRY	idEntries[1]; // the entries for each image
} ICONDIR, *LPICONDIR;


class CIconExtractor
{

public:

	CIconExtractor();
	
	HICON ExtractIconFromExeEx(LPCSTR SourceEXE, UINT IconIndex, HBITMAP &destBitmap, int &maxIconSize);
	HICON ExtractIconFromIcoEx(LPCSTR SourceICO, HBITMAP &destBitmap, int &maxIconSize);

protected:

	BOOL AddResource(LPCTSTR lpszType, LPTSTR lpszName); 

private:

	HICON MakeIconFromResource( LPICONIMAGE lpIcon );

	BOOL AdjustIconImagePointers( LPICONIMAGE lpImage );

	LPSTR FindDIBBits (LPSTR lpbi);
	WORD DIBNumColors (LPSTR lpbi);
	WORD PaletteSize (LPSTR lpbi);
	DWORD BytesPerLine( LPBITMAPINFOHEADER lpBMIH );
	UINT ReadICOHeader( HANDLE hFile );

	ULONG		 m_GroupIconID[MAX_ICONS];
	CStr		 m_GroupIconName[MAX_ICONS];
	
	UINT 		 m_GKounter;
	UINT         m_GNameKounter;

	BOOL static CALLBACK EnumResNameProc( HMODULE hModule,		// module handle
											LPCTSTR lpszType,	// resource type
											LPTSTR lpszName,	// resource name
											LONG_PTR lParam);   // application-defined parameter

};

#endif